--- php-5.2.6.orig/ext/ftp/ftp.c	Mon Dec 31 15:20:06 2007
+++ php-5.2.6/ext/ftp/ftp.c	Sun Sep  7 10:10:39 2008
@@ -242,6 +242,7 @@
 ftp_login(ftpbuf_t *ftp, const char *user, const char *pass TSRMLS_DC)
 {
 #if HAVE_OPENSSL_EXT
+	int	errcode;
 	SSL_CTX	*ctx = NULL;
 #endif
 	if (ftp == NULL) {
@@ -290,11 +291,23 @@
 
 		SSL_set_fd(ftp->ssl_handle, ftp->fd);
 
-		if (SSL_connect(ftp->ssl_handle) <= 0) {
-			php_error_docref(NULL TSRMLS_CC, E_WARNING, "SSL/TLS handshake failed");
-			SSL_shutdown(ftp->ssl_handle);
-			return 0;
-		}
+		do {
+			errcode = SSL_connect(ftp->ssl_handle);
+			switch (SSL_get_error(ftp->ssl_handle, errcode)) {
+				case SSL_ERROR_NONE:
+					errcode = 1;
+					break;
+				case SSL_ERROR_WANT_WRITE:
+				case SSL_ERROR_WANT_READ:
+				case SSL_ERROR_WANT_X509_LOOKUP:
+					errcode = 0;
+					break;
+				default:
+					php_error_docref(NULL TSRMLS_CC, E_WARNING, "SSL/TLS handshake failed");
+					SSL_shutdown(ftp->ssl_handle);
+					return 0;
+			}
+		} while (errcode == 0 && !SSL_is_init_finished(ftp->ssl_handle));
 
 		ftp->ssl_active = 1;
 
@@ -1272,6 +1285,9 @@
 my_recv(ftpbuf_t *ftp, php_socket_t s, void *buf, size_t len)
 {
 	int		n, nr_bytes;
+#if HAVE_OPENSSL_EXT
+	int		no_error;
+#endif
 
 	n = php_pollfd_for_ms(s, PHP_POLLREADABLE, ftp->timeout_sec * 1000);
 	if (n < 1) {
@@ -1287,7 +1303,25 @@
 	if (ftp->use_ssl && ftp->fd == s && ftp->ssl_active) {
 		nr_bytes = SSL_read(ftp->ssl_handle, buf, len);
 	} else if (ftp->use_ssl && ftp->fd != s && ftp->use_ssl_for_data && ftp->data->ssl_active) {	
-		nr_bytes = SSL_read(ftp->data->ssl_handle, buf, len);
+		no_error = 1; // there is no error
+		do {
+			nr_bytes = SSL_read(ftp->data->ssl_handle, buf, len);
+			switch (SSL_get_error(ftp->data->ssl_handle, nr_bytes)) {
+				case SSL_ERROR_NONE:
+				case SSL_ERROR_ZERO_RETURN:
+				case SSL_ERROR_WANT_WRITE:
+				case SSL_ERROR_WANT_X509_LOOKUP:
+				case SSL_ERROR_WANT_READ:
+					break;
+				case SSL_ERROR_WANT_CONNECT:
+				case SSL_ERROR_WANT_ACCEPT:
+				case SSL_ERROR_SSL:            
+				case SSL_ERROR_SYSCALL:
+				default:
+					no_error = 0; //major error, get out
+					break;
+			}
+		} while (nr_bytes < 0 && no_error);
 	} else {
 #endif
 		nr_bytes = recv(s, buf, len, 0);
@@ -1491,6 +1525,7 @@
 	socklen_t			size;
 
 #if HAVE_OPENSSL_EXT
+	int		errcode;
 	SSL_CTX		*ctx;
 #endif
 
@@ -1534,11 +1569,23 @@
 			SSL_copy_session_id(data->ssl_handle, ftp->ssl_handle);
 		}
 			
-		if (SSL_connect(data->ssl_handle) <= 0) {
-			php_error_docref(NULL TSRMLS_CC, E_WARNING, "data_accept: SSL/TLS handshake failed");
-			SSL_shutdown(data->ssl_handle);
-			return 0;
-		}
+		do {
+			errcode = SSL_connect(data->ssl_handle);
+			switch (SSL_get_error(data->ssl_handle, errcode)) {
+				case SSL_ERROR_NONE:
+					errcode = 1;
+					break;
+				case SSL_ERROR_WANT_WRITE:
+				case SSL_ERROR_WANT_READ:
+				case SSL_ERROR_WANT_X509_LOOKUP:
+					errcode = 0;
+					break;
+				default:
+					php_error_docref(NULL TSRMLS_CC, E_WARNING, "data_accept: SSL/TLS handshake failed");
+					SSL_shutdown(data->ssl_handle);
+					return 0;
+			}
+		} while (errcode == 0 && !SSL_is_init_finished(data->ssl_handle));
 			
 		data->ssl_active = 1;
 	}	
